/***************************************************************************
 *   Copyright (C) 2004, 2005 Thomas Nagy                                  *
 *   tnagy2^8@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qpoint.h>
#include <qevent.h>
#include <qpen.h>
#include <qcursor.h>
#include <qfont.h>
#include <qpixmap.h>
#include <qwmatrix.h>

#include <kmessagebox.h>
#include <klocale.h>
#include <kdebug.h>
#include <kfontdialog.h>
#include <kcolordialog.h>

#include "DCanvasTip.h"
#include "DCanvasItem.h"
#include "DCanvasPopup.h"
#include "DCanvasView.h"

#include "settings.h"
#include "aux.h"
#include "datastruct/DDataItem.h"
#include "datastruct/DDataControl.h"

#include "DCanvasPos.h"

// z position for link threads
static const int s_linez = 50;

DCanvasView::DCanvasView(QWidget* parent, const char*) : QCanvasView(parent), DGuiView(parent) 
{
	m_pressed       = false;
	m_multiSelect   = false;
	m_ctrlIsPressed = false;

	m_menu          = NULL;
	m_startItem     = NULL;
	m_selectedlink  = NULL;
	m_selectedref   = NULL;

	m_lastitemtoselect = DItem::NOITEM;

	m_tooltip = new DCanvasTip(this);
	m_currentAction = DCanvasView::act_point;
	setCursor(ArrowCursor);

	m_itemtosort = DItem::NOITEM;
	m_sortcount = 0;

	setCanvas(new QCanvas());
	canvas()->setAdvancePeriod(40);

	// rubber line for link action
	m_rubberLine = new QCanvasLine(canvas());
	m_rubberLine->setPoints(0,0, 0,0);
	m_rubberLine->setZ( s_linez );
	m_rubberLine->show();

	// rubber band for select action
	m_rubberRect = new QCanvasRectangle( canvas() );
	m_rubberRect->setSize(10, 10);
	m_rubberRect->move(-10,-10);
	m_rubberRect->show();
	m_rubberRect->setZ(20);

	// DEBUG TODO REMOVE
	/* QWMatrix m;
	   m.scale(0.6, 0.6);
	   setWorldMatrix(m);*/

	// focus on the center of the map newly created
	scrollBy(canvas()->width()/2, canvas()->height()/2);

	setDragAutoScroll( true );
}

DCanvasView::~DCanvasView()
{
	//kdWarning()<<"DCanvasView::~DCanvasView()"<<endl;
	unplug();

	delete m_tooltip;
	delete m_rubberRect;
	delete m_rubberLine;

	// Deleting all items from the canvas properly is tricky
	// We only delete the DCanvasItem because they will remove
	// their own canvas items 

	// an infinite loop here means that there is a bug.
	int count=0;
	while (true)
	{
		QCanvasItemList lst = canvas()->allItems();
		if (lst.count() == 0)
			break;

		for (int i=0; i<lst.count(); i++)
		{
			if (lst[i]->rtti() != DCanvasItem::Rtti_DCanvasItem)
				continue;
			delete (DCanvasItem*) lst[i];
			break;
		}
		if (count > 100000)
		{
			kdWarning() <<"BUG in DCanvasView destructor"<<endl;
			break;
		}
		count++;
	}

	canvas()->setAdvancePeriod(-1);
	delete canvas();
}

void DCanvasView::focusOnRoot()
{
	centerOnObject( dataTree()->rootID() );
	dataTree()->setItemSelected( dataTree()->rootID(), NULL );
}

DCanvasItem* DCanvasView::canvasItem( int id )
{
	if (id == DItem::NOITEM) return NULL;
	return (DCanvasItem*) Item( id );
}

void DCanvasView::contentsMousePressEvent(QMouseEvent* me)
{
	setFocus();

	QPoint p = inverseWorldMatrix().map(me->pos());

	// remove the context menu
	if (m_menu != NULL)
	{
		m_menu->hide();
		disconnect(m_menu, SIGNAL(activated(int)), this, SLOT(popupMenuSel(int)));
		delete m_menu;
		m_menu = NULL;
	}

	if (me->button() == LeftButton)
		m_pressed = true;

	m_moveStart = p;

	int newitem = locateItem( me->pos() );

	// if the user has selected some items previously, let him move them directly
	if ( m_justSelectedSomething )
	{
		m_justSelectedSomething = false;

		if (me->button() != RightButton)
		{
			if ( m_selectedList.count() > 0 && m_selectedList.contains(newitem) )
			{
				QValueList<int>::iterator end = m_selectedList.end();
				for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
				{
					canvasItem(*m_it)->setOffset( p );
				}

				m_justSelectedSomething = false;
				m_canSelect = false;
				return;
			}
		}
	}

	// set the rubberband for selecting items
	m_canSelect = false;
	if (newitem == DItem::NOITEM)
		m_canSelect = true;

	//
	if (me->button() == LeftButton)
	{
		if (act_point == m_currentAction)
		{
			if ( ! m_multiSelect)
				deselectAll();

			if ( newitem != DItem::NOITEM )
			{
				canvasItem(newitem)->setOffset( p );
				if (m_selectedList.contains(newitem))
				{
					canvasItem(newitem)->setSelected(false);
					m_selectedList.remove(newitem);
				}
				else
				{
					addItemToSelection( newitem );
				}

				// Use this to debug the structure of the tree
				//dataTree()->debugItem(newitem);
			}

			if ( m_multiSelect )
			{
				QValueList<int>::iterator end = m_selectedList.end();
				for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
				{
					canvasItem(*m_it)->setOffset( p );
				}
			}
		}
		else if (act_link == m_currentAction)
		{
			QPen pen;
			pen.setWidth( 2 );

			if (!m_ctrlIsPressed)
			{
				pen.setColor( dataTree()->m_col_link );
				pen.setStyle( Qt::SolidLine );
				m_rubberLine->setPen(pen);

				if (m_startItem)
					m_startItem->setSelected(false);

				m_startItem = canvasItem( newitem );

				if (m_startItem)
					m_startItem->setSelected(true);
			}
			else
			{
				pen.setColor( dataTree()->m_col_ref );
				pen.setStyle( Qt::DashLine );
				m_rubberLine->setPen(pen);

				if (m_startItem)
					m_startItem->setSelected(false);

				m_startItem = canvasItem( newitem );

				if (m_startItem)
					m_startItem->setSelected(true);
			}
		}
		else if (act_sort == m_currentAction)
		{
			int child = locateItemPos( me->pos() );
			deselectAll();

			if (child != DItem::NOITEM)
			{
				// set the position of a child
				// kdWarning()<<"sort child on mouse press "<<child<<endl;
				dataItem(m_itemtosort)->setChildIdx(child, m_sortcount);

				m_sortcount++;
				if (m_sortcount >= dataItem(m_itemtosort)->countChildren())
					m_sortcount = 0;

				dataTree()->notifyChildChange(m_itemtosort);
				dataTree()->notifyChildren(m_itemtosort);
			}
			else
			{
				// select another parent for sorting his children
				m_sortcount = 0;
				if (m_itemtosort != DItem::NOITEM)
				{
					//canvasItem(m_itemtosort)->setSelected(false);
					for (int i=0; i<dataItem(m_itemtosort)->countChildren(); i++)
					{
						canvasItem(dataItem(m_itemtosort)->childNum(i))->setDisplayIdx(false);
					}
				}

				m_itemtosort = newitem;

				if (m_itemtosort != DItem::NOITEM)
				{
					//canvasItem(m_itemtosort)->setSelected(true);
					addItemToSelection( m_itemtosort );
					dataTree()->setItemSelected( m_selectedList[0], this );

					for (int i=0; i<dataItem(m_itemtosort)->countChildren(); i++)
					{
						canvasItem(dataItem(m_itemtosort)->childNum(i))->setDisplayIdx(true);
					}
				}
				else
				{
					dataTree()->setItemSelected( DItem::NOITEM, this );
				}
			}
			canvas()->update();
		}
		else if (act_scroll == m_currentAction)
		{
			deselectAll();
			if ( newitem != DItem::NOITEM )
			{
				//canvasItem(newitem)->setSelected(true);
				addItemToSelection( newitem );
				dataTree()->setItemSelected( m_selectedList[0], this );
			}
			else
			{
				dataTree()->setItemSelected( DItem::NOITEM, this );
			}
		}
	}

	if (me->button() == RightButton)
	{
		if ( ! m_multiSelect && !m_selectedList.contains(newitem) )
			deselectAll();

		if ( newitem != DItem::NOITEM )
		{
			canvasItem(newitem)->setOffset( p );
			if (!m_selectedList.contains(newitem))
				addItemToSelection( newitem );
		}
		else
		{
			deselectAll();
		}

		if (m_readWrite)
		{
			m_selectedlink = NULL;
			m_selectedref  = NULL;
			QCanvasItemList l = canvas()->collisions( p );
			QCanvasItemList::iterator it;
			QCanvasItemList::iterator end = l.end();
			for ( it = l.begin(); it != end; ++it  )
			{
				if ( (*it)->rtti() == DCanvasRef::Rtti_DCanvasRef )
				{
					m_selectedref = (DCanvasRef*) (*it);
					break;
				}
				else if ( (*it)->rtti() == DCanvasLink::Rtti_DCanvasLink )
				{
					m_selectedlink = (DCanvasLink*) (*it);
					break;
				}
			}

			m_menu = new DCanvasPopup(this);
			m_menu->popup(me->globalPos());
			connect(m_menu, SIGNAL(activated(int)), this, SLOT(popupMenuSel(int)));
		}
	}
}

void DCanvasView::addItemToSelection( int id )
{
	if (id == DItem::NOITEM)
		return;

	if (!isRegistered(id))
		return;

	if (m_selectedList.count() == 0)
		m_lastitemtoselect = id;

	canvasItem(id)->setSelected(true);

	if (m_selectedList.contains(id) == 0)
		m_selectedList.push_back(id);
}

void DCanvasView::contentsMouseDoubleClickEvent(QMouseEvent* me)
{
	deselectAll();

	QPoint p = inverseWorldMatrix().map( me->pos() );

	int newitem = locateItem( p );
	if (m_readWrite)
	{
		if (m_ctrlIsPressed)
		{
			if (newitem != DItem::NOITEM)
			{
				dataTree()->setOrphan( newitem );
			}
			else
			{
				QRect r(p.x()-5, p.y()-5, 10, 10);

				QCanvasItemList l = canvas()->collisions( r );
				QCanvasItemList::iterator it;
				QCanvasItemList::iterator end = l.end();
				for ( it = l.begin(); it != end; ++it  )
				{
					if ( (*it)->rtti() == DCanvasLink::Rtti_DCanvasLink )
					{
						DCanvasLink* link = (DCanvasLink*) (*it);
						dataTree()->setOrphan( link->IdOrig() );
						break;
					}
					else if ( (*it)->rtti() == DCanvasRef::Rtti_DCanvasRef )
					{
						DCanvasRef* ref = (DCanvasRef*) (*it);
						dataTree()->requestRefChange( ref->IdOrig(), ref->IdDest(), false );
						break;
					}
				}
			}
		}
		else
		{
			dataTree()->createItem( newitem );
		}
	}
}

void DCanvasView::contentsMouseMoveEvent(QMouseEvent* me)
{
	QPoint p = inverseWorldMatrix().map(me->pos());
	if (m_pressed && m_currentAction == act_point)
	{
		if (m_canSelect)
		{
			// draw the rubberband for selecting many items at once

			// move the rubberband
			m_rubberRect->move( m_moveStart.x(), m_moveStart.y() );
			m_rubberRect->setSize( p.x()-m_moveStart.x(), p.y()-m_moveStart.y());

			// set 'selected' all items inside the rubberband
			QCanvasItemList l = canvas()->collisions( m_rubberRect->rect() );
			QCanvasItemList::iterator it;

			// remove the previous selection
			QValueList<int>::iterator endl = m_selectedList.end();
			for ( m_it = m_selectedList.begin(); m_it != endl; ++m_it )
				canvasItem(*m_it)->setSelected( false );
			m_selectedList.clear();

			// add the ids of the items to the selection
			QCanvasItemList::iterator end = l.end();
			for ( it = l.begin(); it != end; ++it  )
			{
				if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem) continue;

				DCanvasItem *citem = (DCanvasItem*) (*it);
				if (!citem->isSelected()) citem->setSelected( true );

				addItemToSelection( citem->Id() );
			}

			m_justSelectedSomething = true;
		}
		else if (!m_multiSelect && m_readWrite)
		{
			// do not move items while doing a selection (annoying behaviour)

			// move the selected items
			QValueList<int>::iterator end = m_selectedList.end();
			for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
			{
				DCanvasItem *citem = canvasItem(*m_it);

				if (!citem)
					return;

				if (!citem->isSelected())
					citem->setSelected(true);

				citem->moveOffset(p);

				DDataItem *item = dataItem( *m_it );
				updateObjLinks( *m_it );
			}

			// the document has changed
			dataTree()->docChanged();
		}
	}

	// in link mode, draw the thread for linking items
	if (m_currentAction == act_link)
	{
		if (m_rubberLine)
			m_rubberLine->setPoints( m_moveStart.x(), m_moveStart.y(), p.x(), p.y());
	}

	// in scroll mode .. well, scroll :)
	if (m_currentAction == act_scroll)
	{
		scrollBy( m_moveStart.x() - p.x(), m_moveStart.y() - p.y() );
	}
}

void DCanvasView::contentsMouseReleaseEvent(QMouseEvent* me)
{
	m_pressed = false;

	m_rubberRect->move( -10, -10 );
	m_rubberRect->setSize( 10, 10);

	canvas()->setAllChanged();

	int newitem = locateItem( me->pos() );

	if (m_currentAction == act_link && m_readWrite)
	{
		m_rubberLine->setPoints(0, 0, 0, 0);

		if (m_startItem)
		{
			m_startItem->setSelected(false);
			int id = locateItem(me->pos());

			if (id != DItem::NOITEM)
			{
				if (m_startItem->Id() != id)
				{
					if (!m_ctrlIsPressed)
						dataTree()->linkItems( m_startItem->Id(), id );
					else
						dataTree()->requestRefChange( m_startItem->Id(), id, true );
				}
				else
				{
					deselectAll();
					addItemToSelection(id);
				}
				//updateItem(id);
			}
			updateItem(m_startItem->Id());
		}
		m_startItem = NULL;
	}

	if (m_currentAction == act_point)
	{
		if (m_lastitem != DItem::NOITEM && newitem != DItem::NOITEM && m_ctrlIsPressed)
		{
			dataTree()->linkItems( m_lastitem, newitem );
		}
		m_lastitem = newitem;
	}

	if (oneItemIsSelected())
		dataTree()->setItemSelected( m_selectedList[0], this );
	else
		dataTree()->setItemSelected( DItem::NOITEM, this );
}

void DCanvasView::contentsWheelEvent( QWheelEvent* e )
{
	if (!m_multiSelect)
	{
		e->ignore();
		return;
	}

	int incval = 1;
	if (e->delta() < 0)
	{
		incval = -1;
	}

	if (itemsAreSelected())
	{
		QValueList<int>::iterator end = m_selectedList.end();
		for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
		{
			DDataItem *item = dataItem(*m_it);
			item->incrementFontSize(incval);
			canvasItem(*m_it)->updatePos();
		}
	}
	else
	{
		int id = locateItem(e->pos());
		if (id != DItem::NOITEM)
		{
			DDataItem *item = dataItem(id);
			item->incrementFontSize(incval);
			canvasItem(id)->updatePos();
		}
	}
}

int DCanvasView::locateItem( const QPoint& p )
{
	DCanvasItem* item = NULL;
	DCanvasItem* obj  = NULL;

	QCanvasItemList l = canvas()->collisions( inverseWorldMatrix().map(p) );
	QCanvasItemList::iterator it;

	double z=-1;

	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
			continue;

		obj = (DCanvasItem*) (*it);

		if ( obj->z() > z )
		{
			z = obj->z();
			item = obj;
		}
	}

	if (!item)
		return DItem::NOITEM;

	return item->Id();
}

int DCanvasView::locateItemPos( const QPoint& p)
{
	DCanvasPos* pos = NULL;
	DCanvasPos* obj = NULL;
	QCanvasItemList l = canvas()->collisions( inverseWorldMatrix().map(p) );
	QCanvasItemList::iterator it;

	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != Rtti_DCanvasPos)
			continue;

		obj = (DCanvasPos*) (*it);
		break;
	}

	pos = obj;

	if (!pos)
		return DItem::NOITEM;

	return pos->Id();
}

void DCanvasView::moveSelObjectsUp()
{
	kdDebug()<<"DCanvasView::moveSelObjectsUp"<<endl;
	for (int i=0; i<m_selectedList.count(); i++)
	{
		DCanvasItem *item = canvasItem( m_selectedList[i] );
		item->move(item->x(), item->y() - 30);
		updateObjLinks( m_selectedList[i] );
	}
}

void DCanvasView::moveSelObjectsDown()
{
	kdDebug()<<"DCanvasView::moveSelObjectsDown"<<endl;
	for (int i=0; i<m_selectedList.count(); i++)
	{
		DCanvasItem *item = canvasItem( m_selectedList[i] );
		item->move(item->x(), item->y() + 30);
		updateObjLinks( m_selectedList[i] );
	}
}

void DCanvasView::moveSelObjectsLeft()
{
	kdDebug()<<"DCanvasView::moveSelObjectsLeft"<<endl;
	for (int i=0; i<m_selectedList.count(); i++)
	{
		DCanvasItem *item = canvasItem( m_selectedList[i] );
		item->move(item->x() - 30, item->y());
		updateObjLinks( m_selectedList[i] );
	}
}

void DCanvasView::moveSelObjectsRight()
{
	kdDebug()<<"DCanvasView::moveSelObjectsRight"<<endl;
	for (int i=0; i<m_selectedList.count(); i++)
	{
		DCanvasItem *item = canvasItem( m_selectedList[i] );
		item->move(item->x() + 30, item->y());
		updateObjLinks( m_selectedList[i] );
	}
}


void DCanvasView::selectObjUp()
{
	if (! oneItemIsSelected()) return;

	DDataItem *item = dataItem( m_selectedList[0] );
	if (item->Parent() != DItem::NOITEM)
	{
		deselectAll();
		int newstuff = item->Parent();
		addItemToSelection( newstuff );
		makeObjectVisible( newstuff );
		dataTree()->setItemSelected( newstuff, this );
	}
}

void DCanvasView::selectObjDown()
{
	if (!oneItemIsSelected()) return;

	DDataItem *item = dataItem( m_selectedList[0] );
	if (item->countChildren() > 0)
	{
		int newstuff = DItem::NOITEM;
		int lastchld = item->m_lastchild;
		if (lastchld < item->countChildren())
			newstuff = item->childNum(lastchld);
		else
			newstuff = item->childNum(0);


		deselectAll();
		addItemToSelection( newstuff );
		makeObjectVisible( newstuff );
		dataTree()->setItemSelected( newstuff, this );
	}

}

void DCanvasView::selectObjLeft()
{
	if (!oneItemIsSelected()) return;

	DDataItem *item = dataItem( dataItem( m_selectedList[0] )->Parent() );
	if (item)
	{
		int idx = item->childIdx( m_selectedList[0] );
		int newidx = (idx-1+item->countChildren())%item->countChildren();
		item->m_lastchild = newidx;

		int newstuff = item->childNum(newidx);
		deselectAll();
		addItemToSelection( newstuff );
		makeObjectVisible( newstuff );
		dataTree()->setItemSelected( newstuff, this );
	}
}

void DCanvasView::selectObjRight()
{
	if (!oneItemIsSelected()) return;

	DDataItem *item = dataItem( dataItem( m_selectedList[0] )->Parent() );
	if (item)
	{
		int idx = item->childIdx( m_selectedList[0] );
		int newidx = (idx+1)%item->countChildren();
		item->m_lastchild = newidx;

		int newstuff = item->childNum(newidx);
		deselectAll();
		addItemToSelection( newstuff );
		makeObjectVisible( newstuff );
		dataTree()->setItemSelected( newstuff, this );
	}
}

void DCanvasView::keyPressEvent( QKeyEvent* e )
{
	kdDebug()<<"DCanvasView::keyPressEvent"<<endl;
	
	switch (e->key())
	{
		case Qt::Key_Shift:
			// shift key is held pressed
			m_multiSelect = true;
			break;

		case Qt::Key_Delete:
			if (m_readWrite)
				removeSelectedItems();
			break;

		case Qt::Key_Backspace:
			if (oneItemIsSelected())
			{
				DDataItem *item = dataItem( m_selectedList[0] );
				int newsize = item->m_summary.length() - 1;
				if (newsize < 0) newsize = 0;
				item->m_summary.truncate( newsize );
				dataTree()->notifyItemModified( m_selectedList[0] );
			}
			break;

		case Qt::Key_Control:
			m_lastitem = DItem::NOITEM;
			//if (e->state() & ControlButton)
				m_ctrlIsPressed = true;
				e->ignore();
			break;

		case Qt::Key_Up:
			if ( !(e->state() & AltButton) && !(e->state() & ControlButton) && !(e->state() & MetaButton))
				scrollBy(0, -30);
			else e->ignore();
			break;
		case Qt::Key_Down:
			if ( !(e->state() & AltButton) && !(e->state() & ControlButton) && !(e->state() & MetaButton))
				scrollBy(0, 30);
			else e->ignore();
			break;
		case Qt::Key_Left:
			if ( !(e->state() & AltButton) && !(e->state() & ControlButton) && !(e->state() & MetaButton))
				scrollBy(-30, 0);
			else e->ignore();
			break;
		case Qt::Key_Right:
			if ( !(e->state() & AltButton) && !(e->state() & ControlButton) && !(e->state() & MetaButton))
				scrollBy(30, 0);
			else e->ignore();
			break;

		case Key_Tab:
		case Key_Prior:
		case Key_Return:
		case Key_Enter:
		case Key_F16:
		case Key_F18:
		case Key_F20:
		case Key_Direction_L:
		case Key_Direction_R:
			e->ignore();
			break;

		default:

			if (!itemsAreSelected())
			{
				addItemToSelection( m_lastitemtoselect );
			}

			if (oneItemIsSelected() && !( e->state() & ControlButton )
					&& !( e->state() & AltButton )
					&& !( e->state() & MetaButton )
					&& e->ascii() >= 32 )
			{
				QString txt = e->text();
				if (txt.length() > 0)
				{
					DDataItem *item = dataItem( m_selectedList[0] );
					item->m_summary.append( txt );
					dataTree()->notifyItemModified( m_selectedList[0] );
				}
				else e->ignore();
			}
			else e->ignore();

			break;
	}
}

void DCanvasView::keyReleaseEvent( QKeyEvent* e )
{
	// kdWarning()<<"key released"<<endl;
	switch (e->key())
	{
		case Qt::Key_Shift:
			m_multiSelect = false;
			m_justSelectedSomething = true;
			break;

		case Qt::Key_Control:
			m_ctrlIsPressed = false;
			break;

		default:
			e->ignore();
			break;
	}
	//m_previousAction = m_currentAction;
}

void DCanvasView::deselectAll()
{
	QValueList<int>::iterator end = m_selectedList.end();
	for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
	{
		canvasItem(*m_it)->setSelected(false);
	}
	m_selectedList.clear();
}

bool DCanvasView::oneItemIsSelected() const
{
	if (m_selectedList.count() == 1)
		return true;		
	return false;
}

bool DCanvasView::itemsAreSelected() const
{
	return (m_selectedList.count() > 0);
}

void DCanvasView::popupMenuSel(int sel)
{
	switch (sel)
	{
		case DCanvasPopup::e_item:
			if (oneItemIsSelected())
				dataTree()->createItem( m_selectedList[0] );
			else
				dataTree()->createItem();
			break;

		case DCanvasPopup::e_delete:
			removeSelectedItems();
			break;

		case DCanvasPopup::e_discon_parent:
			if (oneItemIsSelected())
				dataTree()->setOrphan( m_selectedList[0] );
			break;

		case DCanvasPopup::e_discon_obj:
			{
				QValueList<int>::iterator end = m_selectedList.end();
				for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
					dataTree()->disconnectItem( *m_it );
			}
			break;

		case DCanvasPopup::e_discon_subtree:
			if (oneItemIsSelected())
				dataTree()->killFamily( m_selectedList[0] );
			break;

		case DCanvasPopup::e_tidy_all:
			tidyCanvas();
			break;

		case DCanvasPopup::e_tidy_clean:
			tidyClean();
			break;

		case DCanvasPopup::e_tidy_unclutter:
			tidyUnclutter();
			break;

		case DCanvasPopup::e_font:
			{
				QFont font;

				if (oneItemIsSelected())
					font = dataItem( m_selectedList[0] )->m_defaultFont;
				else
					font = dataTree()->m_canvasFont;

				int result = KFontDialog::getFont( font );
				if ( result == KFontDialog::Accepted )
				{
					QValueList<int>::iterator end = m_selectedList.end();
					for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
					{
						dataItem(*m_it)->m_defaultFont = font;
						canvasItem(*m_it)->updatePos();
					}
				}
			}
			break;

		case Settings::EnumColorMode::custom_:
			{
				QColor myColor;
				int result = KColorDialog::getColor( myColor );

				if ( result == KColorDialog::Accepted )
				{
					QValueList<int>::iterator end = m_selectedList.end();
					for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
					{
						dataItem(*m_it)->setCustomColors( 
								myColor, 
								QColor( Qt::black ), 
								QColor( Qt::black ) );
						canvasItem(*m_it)->updatePos();
					}
				}
			}
			break;

		case Settings::EnumColorMode::default_:
		case Settings::EnumColorMode::clear_:
		case Settings::EnumColorMode::important_:
		case Settings::EnumColorMode::opposition_:
		case Settings::EnumColorMode::enumeration_:
		case Settings::EnumColorMode::todo_:
		case Settings::EnumColorMode::done_:
			{
				QValueList<int>::iterator end = m_selectedList.end();
				for ( m_it = m_selectedList.begin(); m_it != end; ++m_it )
				{
					dataItem(*m_it)->setColorScheme( sel );
					canvasItem(*m_it)->updatePos();
				}
			}
			break;

		case DCanvasPopup::e_discon_link:
			{
				if (m_selectedref)
				{
					int orig = m_selectedref->IdOrig();
					int dest = m_selectedref->IdDest();
					dataTree()->requestRefChange( orig, dest, false );
					m_selectedref = NULL;
				}
				else if (m_selectedlink)
				{
					dataTree()->setOrphan( m_selectedlink->IdOrig() );
					m_selectedlink = NULL;
				}
				break;
			}

		default:
			break;
	}

	// no multi selection
	m_multiSelect = false;
}

void DCanvasView::settingsChanged()
{
	canvas()->resize(4000, 4000);
	setHScrollBarMode(QScrollView::Auto);
	setVScrollBarMode(QScrollView::Auto);

	QCanvasItemList l = canvas()->allItems();
	QCanvasItemList::iterator it;
	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
			continue;

		DCanvasItem* obj = (DCanvasItem*) (*it);
		obj->updatePos();
		//obj->updateLink();
	}

	QPen pen;
	pen.setStyle( Qt::DotLine );
	if ( dataTree()->m_reverseVideo )
	{
		canvas()->setBackgroundColor( QColor(Qt::black) );
		pen.setColor( QColor( Qt::white ) );
	}
	else
	{
		canvas()->setBackgroundColor( dataTree()->m_col_background );
		pen.setColor( QColor( Qt::black ) );
	}
	m_rubberRect->setPen( pen );
}

void DCanvasView::removeSelectedItems()
{
	disconnect(dataTree(), SIGNAL(itemSelected(int, DGuiView*)),
			this, SLOT( updateSelectedItem(int, DGuiView*)) );

	while ( m_selectedList.count() > 0 )
	{
		dataTree()->removeItem( m_selectedList.front() );
	}

	connect(dataTree(), SIGNAL(itemSelected(int, DGuiView*)),
			this, SLOT( updateSelectedItem(int, DGuiView*)) );
}

// Do not use directly, it is called by the data
void DCanvasView::createItem( int idx )
{
	m_lastitemtoselect = idx;
	DCanvasItem* item = new DCanvasItem(this, idx);

	if (!item)
		return;

	registerItem(item);
	if (dataItem(idx)->x() == 0 && dataItem(idx)->y() == 0)
		item->move( (double) m_moveStart.x(), (double) m_moveStart.y());
	//else
	// kdWarning()<<"x and y are :"<<dataItem(idx)->x()<<" "<<dataItem(idx)->y()<<endl;

	item->show();
}

void DCanvasView::setActionType(ActionType action)
{
	deselectAll();
	dataTree()->setItemSelected( DItem::NOITEM, this );

	m_currentAction = action;
	m_justSelectedSomething = false;

	switch (m_currentAction)
	{
		case act_point:
			setCursor(ArrowCursor);
			break;
		case act_link:
			setCursor(CrossCursor);
			break;
		case act_sort:
			setCursor(PointingHandCursor);
			break;
		case act_scroll:
			setCursor(SizeAllCursor);
			break;
		default:
			break;
	}

	m_rubberRect->move( -10, -10 );
	m_rubberRect->setSize( 10, 10);

	m_rubberLine->setPoints(0,0, 0,0);

	if (m_currentAction == act_link)
	{
		if (Settings::reverseVideo())
			m_rubberLine->setPen( QColor(Qt::white) );
		else
			m_rubberLine->setPen( QColor(Qt::black) );
	}

	if (m_currentAction != act_sort)
	{
		if (m_itemtosort != DItem::NOITEM)
		{
			canvasItem(m_itemtosort)->setSelected(false);
			for (int i=0; i<dataItem(m_itemtosort)->countChildren(); i++)
			{
				canvasItem(dataItem(m_itemtosort)->childNum(i))->setDisplayIdx(false);
			}
		}		      
	}
}

void DCanvasView::updateItem(int id)
{
	DCanvasItem* item = canvasItem(id);
	if (!item)
	{
		//kdWarning()<<"DCanvasView::updateItem unknown item (probably updating a document) "<<id<<endl;
		return;
	}

	//kdWarning()<<"DCanvasView::updateItem "<<id<<" "<<item<<endl;

	item->updatePos();
	//item->updateLink();
}

void DCanvasView::plug()
{
	if (!dataTree())
		return;
	connect(dataTree(), SIGNAL(settingsChanged()), this, SLOT(settingsChanged()) );
	connect(dataTree(), SIGNAL(itemChanged(int)), this, SLOT(updateItem(int)) );
	connect(dataTree(), SIGNAL(itemCreated(int)), this, SLOT(createItem(int)) );
	connect(dataTree(), SIGNAL(itemRemoved(int)), this, SLOT(removeItem(int)) );
	connect(dataTree(), SIGNAL(refChanged(int, int, bool)), 
			this, SLOT(changeRef(int, int, bool)) );
	connect(dataTree(), SIGNAL(itemSelected(int, DGuiView*)), 
			this, SLOT(updateSelectedItem(int, DGuiView*)) );

	// load the document settings
	settingsChanged();
}

void DCanvasView::unplug()
{
	if (!dataTree())
		return;
	disconnect(dataTree(), SIGNAL(settingsChanged()), this, SLOT(settingsChanged()) );
	disconnect(dataTree(), SIGNAL(itemChanged(int)), this, SLOT(updateItem(int)) );
	disconnect(dataTree(), SIGNAL(itemCreated(int)), this, SLOT(createItem(int)) );
	disconnect(dataTree(), SIGNAL(itemRemoved(int)), this, SLOT(removeItem(int)) );
	disconnect(dataTree(), SIGNAL(refChanged(int, int, bool)),
			this, SLOT(changeRef(int, int, bool)) );
	disconnect(dataTree(), SIGNAL(itemSelected(int, DGuiView*)), 
			this, SLOT( updateSelectedItem(int, DGuiView*)) );
}

void DCanvasView::updateSelectedItem(int id, DGuiView* view)
{
	if (view == this)
		return;

	if (id == DItem::NOITEM)
	{
		deselectAll();
	}
	else if (oneItemIsSelected())
	{
		if (m_selectedList[0] != id)
		{
			deselectAll();
			addItemToSelection(id);
		}
	}
	else
	{
		deselectAll();
		addItemToSelection(id);
	}
}

void DCanvasView::removeItem(int id)
{
	unregisterItem(id);
	m_selectedList.remove(id);
	if (m_itemtosort == id)
		m_itemtosort = DItem::NOITEM;

	canvas()->update();
}

void DCanvasView::drawContents( QPainter* p, int cx, int cy, int cw, int ch)
{
	QCanvasView::drawContents(p, cx, cy, cw, ch);
	emit redrawDone();
}

void DCanvasView::tidyClean()
{
	int result = KMessageBox::warningContinueCancel(
			this,
			i18n("KDissert is about to reorganize the whole mindmap using the classic mindmap organization scheme. Is that what you want ?"),
			i18n("Reorganize the mindmap"),
			i18n("Reorganize"),
			"kdissert_clean_reorganization");

	if (result != KMessageBox::Continue)
		return;

	QCanvasItemList l = canvas()->allItems();
	QCanvasItemList::iterator it;
	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
			continue;

		DCanvasItem* item = (DCanvasItem*) (*it);
		DDataItem* data = dataItem( item->Id() );

		if ( data->Parent() == DItem::NOITEM )
		{
			if (data->countChildren() <= 0)
				continue;

			int idxmiddle = 0;
			int cumulated = 0;
			int totalsize = item->sizeH();

			for (int i=0; i<data->countChildren(); i++)
			{
				cumulated += canvasItem( data->childNum(i) )->sizeH();

				if (cumulated > totalsize/2)
				{
					idxmiddle = i;
					if (idxmiddle > 0 && idxmiddle == data->countChildren()-1)
					{
						idxmiddle--;
						cumulated -= canvasItem( data->childNum(i) )->sizeH();
					}
					break;
				}
			}

			int sizeleft = cumulated;
			int sizeright = totalsize - cumulated;

			int orig = 0;

			// process the branches on the left
			orig = data->y() + item->height()/2 + NESTING/2 - sizeleft/2;
			for (int i=0; i<=idxmiddle; i++)
			{
				DCanvasItem* tmpitem = canvasItem( data->childNum(i) );
				tmpitem->setPosition((int) item->x() - BINSPACE - tmpitem->width(), orig, false, false);
				orig += tmpitem->sizeH();
			}

			// process the branches on the right
			// if  orig = data->y() + item->height()/2 + NESTING/2 - sizeright/2; else
			if (Settings::reorgCyclicMap())
				orig = data->y() + item->height()/2 - NESTING/2 + sizeright/2;
			else
				orig = data->y() + item->height()/2 + NESTING/2 - sizeright/2;

			for (int i=idxmiddle+1; i<data->countChildren(); i++)
			{
				DCanvasItem* tmpitem = canvasItem( data->childNum(i) );
				tmpitem->setPosition((int) item->x() + BINSPACE + item->width(), orig, true, Settings::reorgCyclicMap());
				//tmpitem->setPosition((int) item->x() + BINSPACE + item->width(), orig, true, false);
				orig -= tmpitem->sizeH();
			}
		}
	}    
}

void DCanvasView::tidyCanvas()
{
	int result = KMessageBox::warningContinueCancel(
			this,
			i18n("KDissert is about to reorganize the whole mindmap using the smart(tm) reorganization heuristic. It can take a lot of time and give bad results on huge mindmaps. Proceed ?"),
			i18n("Reorganize the mindmap"),
			i18n("Reorganize"),
			"kdissert_smart_reorganization");

	if (result != KMessageBox::Continue)
		return;

	// clear the selection
	QValueList<int>::iterator selend = m_selectedList.end();
	for ( m_it = m_selectedList.begin(); m_it != selend; ++m_it )
	{
		canvasItem(*m_it)->setSelected( false );
	}
	m_selectedList.clear();

	int oldy=0;

	QCanvasItemList l = canvas()->allItems();
	QCanvasItemList::iterator it;
	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
			continue;

		DCanvasItem* item = (DCanvasItem*) (*it);
		DDataItem* data = dataItem( item->Id() );

		if ( data->Parent() == DItem::NOITEM )
		{
			// TODO : the unconnected items are spent way too far
			/*if ( data->countChildren() == 0 )
			  {
			  oldy += 10;
			  item->move(10, oldy+10);
			  oldy += item->height();
			  }*/
			//else
			{
				item->expandSubtree(-2.*PI, 0);
			}
		}
		//item->updatePos();
	}
	tidyUnclutter(); 
}

void DCanvasView::tidyUnclutter()
{
	//kdWarning()<<"unclutter"<<endl;

	QCanvasItemList l = canvas()->allItems();

	bool cluttered = true;
	int count = 0;

	while (cluttered && count < 500)
	{
		//kdWarning()<<"iteration : "<<count<<endl;

		// conditions for stopping
		cluttered = false;
		count++;

		// compute the changes
		QCanvasItemList::iterator it;
		QCanvasItemList::iterator end = l.end();
		for ( it = l.begin(); it != end; ++it  )
		{
			if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
				continue;

			DCanvasItem *item = (DCanvasItem*) *it;
			item->m_vecX = 0;
			item->m_vecY = 0;


			double miditem_x = item->x()+item->width()/2.;
			double miditem_y = item->y()+item->height()/2.;

			// find overlapping items
			QCanvasItemList::iterator tit;
			for ( tit = l.begin(); tit != end; ++tit  )
			{
				//kdWarning() <<"collision"<<endl;
				if ( (*tit)->rtti() != DCanvasItem::Rtti_DCanvasItem )
					continue;

				// to pronounce with a German accent
				DCanvasItem *intruder = (DCanvasItem*) *tit;

				if (item == intruder)
					continue;

				double midintruder_x = intruder->x()+intruder->width()/2.;
				double midintruder_y = intruder->y()+intruder->height()/2.;

				double lenx = (item->width() + intruder->width())/2.;
				double leny = (item->height() + intruder->height())/2.;

				if ( ABS(miditem_x - midintruder_x) > 15+lenx || ABS(miditem_y - midintruder_y) > 15+leny )
					continue;

				cluttered = true;

				if (midintruder_x >= miditem_x)
				{
					item->m_vecX -= 1;
				}
				else
				{
					item->m_vecX += 1;
				}

				if (midintruder_y >= miditem_y)
				{
					item->m_vecY -= 1;
				}
				else
				{
					item->m_vecY += 1;
				}
			}
		}

		// apply the changes
		for ( it = l.begin(); it != end; ++it  )
		{
			if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
				continue;

			DCanvasItem *item = (DCanvasItem*) *it;
			item->move(item->x() + item->m_vecX, item->y() + item->m_vecY);
		}
	}
}

QRect DCanvasView::canvasSize()
{
	double minx = -1, miny = -1, maxx = -1, maxy = -1;

	if ( dataTree()->rootID() == DItem::NOITEM )
	{
		return canvas()->rect();
	}

	DCanvasItem* root = canvasItem(dataTree()->rootID());
	minx = root->x();
	miny = root->y();
	maxx = root->x()+root->width();
	maxy = root->y()+root->height();

	// return the minimum window containing the mindmap
	QCanvasItemList l = canvas()->allItems();
	QCanvasItemList::iterator it;
	QCanvasItemList::iterator end = l.end();
	for ( it = l.begin(); it != end; ++it  )
	{
		if ( (*it)->rtti() != DCanvasItem::Rtti_DCanvasItem )
			continue;

		DCanvasItem* obj = (DCanvasItem*) (*it);

		minx = MIN( minx, obj->x()) ;
		maxx = MAX( maxx, obj->x()+obj->width() );
		miny = MIN( miny, obj->y()) ;
		maxy = MAX( maxy, obj->y()+obj->height() );
	}

	double offset = 20;
	minx -= offset;
	maxx += offset;
	miny -= offset;
	maxy += offset;

	minx = MAX(minx, 0);
	miny = MAX(miny, 0);
	maxx = MIN(maxx, canvas()->width());
	maxy = MIN(maxy, canvas()->height());

	return QRect((int) minx, (int) miny, (int) (maxx-minx), (int) (maxy-miny) );
}

DDataItem* DCanvasView::selectedDataItem()
{
	if (m_selectedList.count() != 1)
		return NULL;
	return dataItem(m_selectedList[0]);
}

void DCanvasView::changeRef(int orig, int dest, bool add)
{
	DCanvasItem *item = canvasItem(orig);
	if (!item)
	{
		kdWarning()<<"BUG DCanvasView::changeRef "<<orig<<endl;
		return;
	}

	if (add)
		item->addRef(dest);
	else
		item->rmRef(dest);
}

void DCanvasView::focusOnNextRoot()
{
	int prev_root = DItem::NOITEM;
	if (oneItemIsSelected())
	{
		// there is already a root selected
		DDataItem *item = dataItem( m_selectedList[0] );
		if ( item->Parent() == DItem::NOITEM )
			prev_root = m_selectedList[0];
	}
	else
	{
		// set the focus on the main root first
		if ( dataTree()->rootID() == DItem::NOITEM ) return;

		deselectAll();
		addItemToSelection( dataTree()->rootID() );
		centerOnObject( dataTree()->rootID() );
		dataTree()->setItemSelected( dataTree()->rootID(), this );
		return;
	}
	
	int rootfound = DItem::NOITEM;
	QMapIterator<int, DItem*> it;
	for (it = m_map.begin(); it != m_map.end(); ++it)
	{
		// cycle and find the previous root, if any
		int id = (*it)->Id();

		if (prev_root == DItem::NOITEM)
		{
			// we have found another root, great
			DDataItem *item = dataItem(id);
			if (item->Parent() == DItem::NOITEM)
			{
				rootfound = id;
				break;
			}

		}
		else if (prev_root == id)
		{
			// we have found the previous root, continue
			prev_root = DItem::NOITEM;
		}
	}

	// still no root ? then get the first one
	if (rootfound == DItem::NOITEM)
	{
		for (it = m_map.begin(); it != m_map.end(); ++it)
		{
			int id = (*it)->Id();
			if (dataItem(id)->Parent() == DItem::NOITEM)
			{
				rootfound = id;
				break;
			}
		}
	}

	if (rootfound != DItem::NOITEM)
	{
		deselectAll();
		addItemToSelection( rootfound );
		centerOnObject( rootfound );
		dataTree()->setItemSelected( rootfound, this );
	}
}

void DCanvasView::centerOnObject( int obj )
{
	DCanvasItem* citem = canvasItem(obj);
	if (!citem) return;

	QRect r = worldMatrix().map( citem->rect() );
	center( (int) (r.x()+r.width()/2), (int) (r.y()+r.height()/2));
}

void DCanvasView::makeObjectVisible( int obj )
{
	DCanvasItem* citem = canvasItem(obj);
	if (!citem) return;

	QRect r = worldMatrix().map( citem->rect() );
	ensureVisible( (int) (r.x()+r.width()/2), (int) (r.y()+r.height()/2), r.width()/2, r.height()/2 );
}

void DCanvasView::updateObjLinks( int obj )
{
	DDataItem *item = dataItem( obj );
	if (!item) return;

	for (int i=0; i<item->countChildren(); i++)
	{
		DCanvasItem *citem = canvasItem( item->childNum(i) );
		citem->updateLink();
	}
}

#include "DCanvasView.moc"
